/*
  Copyright (C) 2003 Grzegorz Godlewski <ggodlewski@arbit.pl>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#if (defined(_MSC_VER) && defined(_WIN32))
#else
using namespace std;
#endif

#include <xercesc/framework/MemBufInputSource.hpp>
#include <xercesc/parsers/SAXParser.hpp>
#include <xercesc/sax/AttributeList.hpp>
#include <xercesc/sax/SAXParseException.hpp>
#include <xercesc/sax/SAXException.hpp>
#include <xercesc/util/TranscodingException.hpp>

#include <math.h>
#include <istream.h>

extern "C"
{
#include <php.h>
#include <zend_globals.h>
}

#if (defined(_MSC_VER) && defined(_WIN32))
#include <config.w32.h>
#else
#include <config.h>
#endif
#include "xerces.h"

void xerces_debug(const char *format, ...)
{
  TSRMLS_FETCH();
  char buffer[1024];
  va_list args;
  va_start(args, format);
  vsnprintf(buffer, sizeof(buffer)-1, format, args);
  va_end(args);
  buffer[sizeof(buffer)-1] = '\0';
  php_printf("Debug: %s<br />\n", buffer);
}

res::res(char *encod)
{
  parser=new SAXParser();
  try
  {
    handler=new MemParseHandlers(this,encod);
  }
  catch (const TranscodingException& e)
  {
    zend_error(E_ERROR,"Xerces exception: %s",e.getMessage());
  }
  parser->setDocumentHandler(handler);
  parser->setErrorHandler(handler);
  object=NULL;
  se_handler=NULL;
  ee_handler=NULL;
  c_handler=NULL;
  df_handler=NULL;

  err_handler=NULL;
  ferr_handler=NULL;
  war_handler=NULL;

  doNamespaces=false;
  doSchema=false;
  strcpy(externalSchemaLocation,"");
  schemaFullChecking=false;
  valScheme=SAXParser::Val_Auto;
}

void res::set_handler(zval **handler, zval **data)
{
  if (Z_TYPE_PP(data) != IS_ARRAY)
  {
    convert_to_string_ex(data);
  }

  if (*handler)
  {
    zval_ptr_dtor(handler);
  }

  zval_add_ref(data);
  *handler = *data;
}

zval *res::call_handler(res *parser, zval *handler, int argc, zval **argv)
{
  TSRMLS_FETCH();

  int argc2=argc;
  if (parser && (handler==NULL))
  {
    handler=parser->df_handler;
    if (argc2>2)argc2=2;
  }

  if (parser && handler)
  {
    zval *retval;
    int i;	
    int result;

    MAKE_STD_ZVAL(retval);
    ZVAL_FALSE(retval);

    result = call_user_function(EG(function_table), &parser->object, handler, retval, argc2, argv TSRMLS_CC);

    if (result == FAILURE)
    {
      zval **method;
      zval **obj;

      if (Z_TYPE_P(handler) == IS_STRING)
      {
        zend_error(E_WARNING, "Unable to call handler %s()", Z_STRVAL_P(handler));
      } else if (zend_hash_index_find(Z_ARRVAL_P(handler), 0, (void **) &obj) == SUCCESS &&
        zend_hash_index_find(Z_ARRVAL_P(handler), 1, (void **) &method) == SUCCESS &&
        Z_TYPE_PP(obj) == IS_OBJECT &&
        Z_TYPE_PP(method) == IS_STRING)
     {
        zend_error(E_WARNING, "Unable to call handler %s::%s()", Z_OBJCE_PP(obj)->name, Z_STRVAL_PP(method));
     } else 
     {
       zend_error(E_WARNING, "Unable to call handler");
     }

     zval_dtor(retval);
     efree(retval);
   }

    for (i = 0; i < argc; i++)
    {
      zval_ptr_dtor(&(argv[i]));
    }

    if (result == FAILURE)
    {
      return NULL;
    } else
    {
      return retval;
    }
  }
  return NULL;
}

res::~res()
{
  if (object)
  {
#ifdef ZEND_ENGINE_2
    zval_ptr_dtor(object);
#endif
    object=NULL;
  }
  delete parser;
  delete handler;
}

void res::set_object(zval *ob)
{
#ifdef ZEND_ENGINE_2
  if (object)
  {
    zval_ptr_dtor(object);
  }
#endif
  object = ob;
#ifdef ZEND_ENGINE_2
  zval_add_ref(object); 
#endif
};

void res::parse(zval *string)
{
  unsigned long duration;
  static const char*  gMemBufId = "php-xerces";

  MemBufInputSource* memBufIS = new MemBufInputSource
    ((const XMLByte*)Z_STRVAL(*string),Z_STRLEN(*string), gMemBufId, false);

  try
  {
    parser->setValidationScheme(valScheme);
    parser->setDoNamespaces(doNamespaces);
    parser->setDoSchema(doSchema);
    parser->setExternalSchemaLocation(externalSchemaLocation);

    parser->setValidationSchemaFullChecking(schemaFullChecking);
    const unsigned long startMillis = XMLPlatformUtils::getCurrentMillis();
    parser->parse(*memBufIS);
    const unsigned long endMillis = XMLPlatformUtils::getCurrentMillis();
    duration = endMillis - startMillis;
  }
  catch (const XMLException& e)
  {
    zend_error(E_ERROR,"Xerces exception: %s",e.getMessage());
  }
}

void destructor_handler(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
  res *r=(res *)rsrc->ptr;
  delete r;
}

static int le_xerces_parser; 

PHP_MINIT_FUNCTION(xerces)
{
  try
  {
    XMLPlatformUtils::Initialize();
  }
  catch (const XMLException& toCatch)
  {
    zend_error(E_ERROR,"Xerces: Error during initialization! Message: %s",toCatch.getMessage());
  }

  le_xerces_parser=zend_register_list_destructors_ex(destructor_handler,NULL,RES_NAME,module_number);

  REGISTER_LONG_CONSTANT("XERCES_OPTION_DO_NAMESPACES", PHP_XERCES_OPTION_DO_NAMESPACES, CONST_CS|CONST_PERSISTENT);
  REGISTER_LONG_CONSTANT("XERCES_OPTION_DO_SCHEMA", PHP_XERCES_OPTION_DO_SCHEMA, CONST_CS|CONST_PERSISTENT);
  REGISTER_LONG_CONSTANT("XERCES_OPTION_SCHEMA_FULL_CHECKING", PHP_XERCES_OPTION_SCHEMA_FULL_CHECKING, CONST_CS|CONST_PERSISTENT);
  REGISTER_LONG_CONSTANT("XERCES_OPTION_VAL_SCHEME", PHP_XERCES_OPTION_VAL_SCHEME, CONST_CS|CONST_PERSISTENT);
  REGISTER_LONG_CONSTANT("XERCES_OPTION_EXTERNAL_SCHEMA_LOCATION", PHP_XERCES_OPTION_EXTERNAL_SCHEMA_LOCATION, CONST_CS | CONST_PERSISTENT);

  return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(xerces)
{
  XMLPlatformUtils::Terminate();
  return 0;
}


ZEND_FUNCTION(xerces_parse)
{
  res *parser;
  zval **pind, **mythis;

  if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(2, &pind, &mythis) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  if (Z_TYPE_PP(mythis) != IS_STRING)
  {
    zend_error(E_WARNING, "Argument 2 has wrong type");
    RETURN_FALSE;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  parser->parse(*mythis);

  RETVAL_TRUE;
}

PHP_FUNCTION(xerces_parse_into_struct)
{
  res *parser;
  zval **pind, **data, **xdata, **info = 0;
  int argc;

  argc = ZEND_NUM_ARGS();
  if (zend_get_parameters_ex(4, &pind, &data, &xdata,&info) == SUCCESS)
  {
    zval_dtor(*info);
    array_init(*info);
  } else if (zend_get_parameters_ex(3, &pind, &data, &xdata) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

//  convert_to_string_ex(data);
//  zval_dtor(*xdata);
//  array_init(*xdata);

//  parser->data = *xdata;
//  if (info)
//    parser->info = *info;
//  parser->level = 0;
//  parser->ltags = emalloc(XML_MAXLEVEL * sizeof(char *));

/*
  XML_SetDefaultHandler(parser->parser, _xml_defaultHandler);
  XML_SetElementHandler(parser->parser, _xml_startElementHandler, _xml_endElementHandler);
  XML_SetCharacterDataHandler(parser->parser, _xml_characterDataHandler);

  ret = XML_Parse(parser->parser, Z_STRVAL_PP(data), Z_STRLEN_PP(data), 1);

  RETVAL_LONG(ret);
*/
}

// HANDLERS
PHP_FUNCTION(xerces_set_element_handler)
{
  res *parser;
  zval **pind, **shdl, **ehdl;
  if (ZEND_NUM_ARGS() != 3 || zend_get_parameters_ex(3, &pind, &shdl, &ehdl) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);
  parser->set_handler(&parser->se_handler, shdl);
  parser->set_handler(&parser->ee_handler, ehdl);
  RETVAL_TRUE;
}

PHP_FUNCTION(xerces_set_default_handler)
{
  res *parser;
  zval **pind, **hdl;
  if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(2, &pind, &hdl) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  parser->set_handler(&parser->df_handler, hdl);
  RETVAL_TRUE;
}

PHP_FUNCTION(xerces_set_character_data_handler)
{
  res *parser;
  zval **pind, **hdl;
  if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(2, &pind, &hdl) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  parser->set_handler(&parser->c_handler, hdl);
  RETVAL_TRUE;
}

PHP_FUNCTION(xerces_set_error_handler)
{
  res *parser;
  zval **pind, **hdl;
  if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(2, &pind, &hdl) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  parser->set_handler(&parser->err_handler, hdl);
  RETVAL_TRUE;
}

PHP_FUNCTION(xerces_set_fatalerror_handler)
{
  res *parser;
  zval **pind, **hdl;
  if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(2, &pind, &hdl) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  parser->set_handler(&parser->ferr_handler, hdl);
  RETVAL_TRUE;
}

PHP_FUNCTION(xerces_set_warning_handler)
{
  res *parser;
  zval **pind, **hdl;
  if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(2, &pind, &hdl) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  parser->set_handler(&parser->war_handler, hdl);
  RETVAL_TRUE;
}

// OPTIONS
ZEND_FUNCTION(xerces_set_object)
{
  res *parser;
  zval **pind, **mythis;

  if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(2, &pind, &mythis) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  if (Z_TYPE_PP(mythis) != IS_OBJECT)
  {
    zend_error(E_WARNING, "Argument 2 has wrong type");
    RETURN_FALSE;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  parser->set_object(*mythis);

  RETVAL_TRUE;
}

PHP_FUNCTION(xerces_parser_set_option)
{
  res *parser;
  zval **pind, **opt, **val;

  if (ZEND_NUM_ARGS() != 3 || zend_get_parameters_ex(3, &pind, &opt, &val) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }
  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  convert_to_long_ex(opt);

  switch (Z_LVAL_PP(opt))
  {
    case PHP_XERCES_OPTION_DO_NAMESPACES:
      convert_to_long_ex(val);
      parser->doNamespaces=Z_LVAL_PP(val)?true:false;
      break;
    case PHP_XERCES_OPTION_DO_SCHEMA:
      convert_to_long_ex(val);
      parser->doSchema=Z_LVAL_PP(val)?true:false;
      break;
    case PHP_XERCES_OPTION_SCHEMA_FULL_CHECKING:
      convert_to_long_ex(val);
      parser->schemaFullChecking=Z_LVAL_PP(val)?true:false;
      break;
    case PHP_XERCES_OPTION_VAL_SCHEME:
      convert_to_string_ex(val);
      if (strcmp(Z_STRVAL_PP(val),"auto") == 0)
      {
        parser->valScheme = SAXParser::Val_Auto;
      } else
      if (strcmp(Z_STRVAL_PP(val),"always") == 0)
      {
        parser->valScheme = SAXParser::Val_Always;
      } else
      if (strcmp(Z_STRVAL_PP(val),"never") == 0)
      {
        parser->valScheme = SAXParser::Val_Never;
      } else
      {
        zend_error(E_WARNING, "Unknown option: \"%s\"", Z_STRVAL_PP(val));
        RETURN_FALSE;
      }
      break;
    case PHP_XERCES_OPTION_EXTERNAL_SCHEMA_LOCATION:
      strcpy(parser->externalSchemaLocation,Z_STRVAL_PP(val));
      break;
    default:
      zend_error(E_WARNING, "Unknown option"); // HACK -> memory fault
      RETURN_FALSE;
      break;
  }
  RETVAL_TRUE;
}

PHP_FUNCTION(xerces_parser_get_option)
{
  res *parser;
  zval **pind, **opt;

  if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(2, &pind, &opt) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }
  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  convert_to_long_ex(opt);

  switch (Z_LVAL_PP(opt))
  {
    case PHP_XERCES_OPTION_DO_NAMESPACES:
      RETURN_LONG(parser->doNamespaces==true?1:0);
      break;
    case PHP_XERCES_OPTION_DO_SCHEMA:
      RETURN_LONG(parser->doSchema?1:0);
      break;
    case PHP_XERCES_OPTION_SCHEMA_FULL_CHECKING:
      RETURN_LONG(parser->schemaFullChecking?1:0);
      break;
    case PHP_XERCES_OPTION_VAL_SCHEME:
      if (parser->valScheme == SAXParser::Val_Auto)
      {
        RETURN_STRING("auto",1);
      } else
      if (parser->valScheme == SAXParser::Val_Always)
      {
        RETURN_STRING("always",1);
      } else
      {
        RETURN_STRING("never",1);
      };
      break;
    case PHP_XERCES_OPTION_EXTERNAL_SCHEMA_LOCATION:
      RETURN_STRING(parser->externalSchemaLocation,1);
      break;
    default:
      zend_error(E_WARNING, "Unknown option");
      RETURN_FALSE;
      break;
  }
  RETVAL_FALSE;
}

///
ZEND_FUNCTION(xerces_parser_create)
{
  zval **par;
  int args;
  char *encoding="ISO-8859-1";

  if ((args=ZEND_NUM_ARGS())>1)
  {
    WRONG_PARAM_COUNT;
  }
  switch (args)
  {
    case 1:
      if (zend_get_parameters_ex(1,&par)!=SUCCESS)
      {
        WRONG_PARAM_COUNT;
      }
      if (Z_TYPE(**par)!=IS_STRING)
      {
        zend_error(E_ERROR,"Xerces: 1st parameter isn't a string");
      }
      convert_to_string_ex(par);
      encoding=Z_STRVAL_PP(par);
      
      break;
    case 0:
      break;
    default:
      WRONG_PARAM_COUNT;
      break;
  }
  
  res *parser=new res(encoding);
  ZEND_REGISTER_RESOURCE(return_value,parser,le_xerces_parser);
  parser->index = Z_LVAL_P(return_value);
}

PHP_FUNCTION(xerces_parser_free)
{
  zval **pind;
  res *parser;

  if (ZEND_NUM_ARGS() != 1 || zend_get_parameters_ex(1, &pind) == FAILURE)
  {
    WRONG_PARAM_COUNT;
  }

  ZEND_FETCH_RESOURCE(parser,res *,pind, -1, RES_NAME, le_xerces_parser);

  if (zend_list_delete(parser->index) == FAILURE)
  {
    RETURN_FALSE;
  }
  RETVAL_TRUE;
}

zend_function_entry xerces_functions[]=
{
  ZEND_FE(xerces_set_object,NULL)
  ZEND_FE(xerces_set_element_handler,NULL)
  ZEND_FE(xerces_set_character_data_handler,NULL)
  ZEND_FE(xerces_set_default_handler,NULL)

  ZEND_FE(xerces_set_error_handler,NULL)
  ZEND_FE(xerces_set_fatalerror_handler,NULL)
  ZEND_FE(xerces_set_warning_handler,NULL)

  ZEND_FE(xerces_parse,NULL)
  ZEND_FE(xerces_parse_into_struct,NULL)

  ZEND_FE(xerces_parser_get_option,NULL)
  ZEND_FE(xerces_parser_set_option,NULL)

  ZEND_FE(xerces_parser_create,NULL)
  ZEND_FE(xerces_parser_free,NULL)

  {NULL,NULL,NULL}
};

zend_module_entry xerces_module_entry=
{
  STANDARD_MODULE_HEADER,
  "xerces",
  xerces_functions,
  PHP_MINIT(xerces),PHP_MSHUTDOWN(xerces),NULL,
  NULL,NULL,
  NO_VERSION_YET,
  STANDARD_MODULE_PROPERTIES
};

#if COMPILE_DL_XERCES
BEGIN_EXTERN_C()
ZEND_GET_MODULE(xerces)
END_EXTERN_C()
#endif

static zval *_xml_resource_zval(long value)
{
  zval *ret;
  TSRMLS_FETCH();
  MAKE_STD_ZVAL(ret);
  Z_TYPE_P(ret) = IS_RESOURCE;
  Z_LVAL_P(ret) = value;
  zend_list_addref(value);
  return ret;
}

static zval *_xml_string_zval(const char *str)
{
  zval *ret;
  int len = strlen(str);
  MAKE_STD_ZVAL(ret);
  Z_TYPE_P(ret) = IS_STRING;
  Z_STRLEN_P(ret) = len;
  Z_STRVAL_P(ret) = estrndup(str, len);
  return ret;
}

static zval *_xml_long_zval(const int i)
{
  zval *ret;
  MAKE_STD_ZVAL(ret);
  Z_TYPE_P(ret) = IS_LONG;
  Z_LVAL_P(ret) = i;
  return ret;
}

void MemParseHandlers::startElement(const XMLCh* const name, AttributeList& attributes)
{
  char att[0x10000], val[0x10000];
  zval *retval, *args[3];

  if (parser->se_handler || parser->df_handler)
  {
    args[0] = _xml_resource_zval(parser->index);
    membuf=(char *)emalloc(1);
    membuf[0]=0;
    fFormatter << XMLFormatter::NoEscapes << name;
    args[1] = _xml_string_zval(membuf);
    MAKE_STD_ZVAL(args[2]);
    array_init(args[2]);

    unsigned int len = attributes.getLength();
    for (unsigned int index = 0; index < len; index++)
    {
      membuf[0]=0;
      fFormatter << XMLFormatter::NoEscapes << attributes.getName(index);
      strncpy(att,membuf,0x10000);
      membuf[0]=0;
      fFormatter << XMLFormatter::NoEscapes << attributes.getValue(index);
      strncpy(val,membuf,0x10000);
      add_assoc_stringl(args[2], att, val, strlen(val), 1);
    }
    if ((retval = parser->call_handler(parser, parser->se_handler, 3, args)))
    {
      zval_dtor(retval);
      efree(retval);
    }
    efree(membuf);
  } 
}

void MemParseHandlers::endElement(const XMLCh* const name)
{
  zval *retval, *args[2];

  if (parser->ee_handler || parser->df_handler)
  {
    args[0] = _xml_resource_zval(parser->index);
    membuf=(char *)emalloc(1);
    membuf[0]=0;
    fFormatter << XMLFormatter::NoEscapes << name;
    args[1] = _xml_string_zval(membuf);

    if ((retval = parser->call_handler(parser, parser->ee_handler, 2, args)))
    {
      zval_dtor(retval);
      efree(retval);
    }
    efree(membuf);
  } 
}

void MemParseHandlers::characters(const XMLCh* const chars, const unsigned int length)
{
  zval *retval, *args[2];

  if (parser->c_handler || parser->df_handler)
  {
    args[0] = _xml_resource_zval(parser->index);
    membuf=(char *)emalloc(1);
    membuf[0]=0;
    fFormatter << XMLFormatter::NoEscapes << chars;
    args[1] = _xml_string_zval(membuf);

    if ((retval = parser->call_handler(parser, parser->c_handler, 2, args)))
    {
      zval_dtor(retval);
      efree(retval);
    }
    efree(membuf);
  } 
}

void MemParseHandlers::ignorableWhitespace(const XMLCh* const chars, const unsigned int length)
{
}

void MemParseHandlers::resetDocument()
{
}

void MemParseHandlers::error(const SAXParseException& e)
{
  zval *retval, *args[4];

  if (parser->err_handler || parser->df_handler)
  {
    args[0] = _xml_resource_zval(parser->index);
    args[1] = _xml_long_zval(e.getLineNumber());
    args[2] = _xml_long_zval(e.getColumnNumber());
    membuf=(char *)emalloc(1);
    membuf[0]=0;
    fFormatter << XMLFormatter::NoEscapes << e.getMessage();
    args[3] = _xml_string_zval(membuf);
    efree(membuf);

    if ((retval = parser->call_handler(parser, parser->err_handler, 4, args)))
    {
      zval_dtor(retval);
      efree(retval);
    }
  } 
}

void MemParseHandlers::fatalError(const SAXParseException& e)
{
  zval *retval, *args[4];

  if (parser->ferr_handler || parser->df_handler)
  {
    args[0] = _xml_resource_zval(parser->index);
    args[1] = _xml_long_zval(e.getLineNumber());
    args[2] = _xml_long_zval(e.getColumnNumber());
    membuf=(char *)emalloc(1);
    membuf[0]=0;
    fFormatter << XMLFormatter::NoEscapes << e.getMessage();
    args[3] = _xml_string_zval(membuf);
    efree(membuf);

    if ((retval = parser->call_handler(parser, parser->ferr_handler, 4, args)))
    {
      zval_dtor(retval);
      efree(retval);
    }
  } 
}

void MemParseHandlers::warning(const SAXParseException& e)
{
  zval *retval, *args[3];

  if (parser->war_handler || parser->df_handler)
  {
    args[0] = _xml_resource_zval(parser->index);
    args[1] = _xml_long_zval(e.getLineNumber());
    args[2] = _xml_long_zval(e.getColumnNumber());
    membuf=(char *)emalloc(1);
    membuf[0]=0;
    fFormatter << XMLFormatter::NoEscapes << e.getMessage();
    args[3] = _xml_string_zval(membuf);
    efree(membuf);

    if ((retval = parser->call_handler(parser, parser->war_handler, 4, args)))
    {
      zval_dtor(retval);
      efree(retval);
    }
  } 
}

MemParseHandlers::MemParseHandlers(res *par,char *encod):
  fFormatter(encod, 0, this, XMLFormatter::NoEscapes, XMLFormatter::UnRep_CharRef)
{
  parser=par;
}

MemParseHandlers::~MemParseHandlers()
{
}
